
from os import PathLike
import os
from typing import Union
from nutree import Tree, Node
from pathlib import Path
from omoospace.exceptions import InvalidError, NotFoundError, NotIncludeError
from omoospace.types import Structure, SubspaceInfo
from omoospace.ui import Grid, Instruction
from omoospace.utils import format_name, is_subpath


class Directory():
    def __init__(
        self,
        path: Path,
        is_subspace: bool = False,
    ):
        self.path = path
        self.is_subspace = is_subspace

    def __repr__(self):
        return self.path.name

    @property
    def ui_name(self):
        if self.is_subspace:
            icon = "🎯"
        else:
            icon = "📂"
        name_str = "[link=%s]%s[/link]" % (self.path, self.path.name)
        return "%s %s" % (icon, name_str)


class DirectoryTree(Tree):

    def __init__(self, structure: Structure = None, search_dir: PathLike = None):
        self.root_path = None
        super().__init__()
        if structure:
            self.from_structure(structure)
        if search_dir:
            self.from_dir(search_dir)

    def from_structure(
        self,
        structure: Structure,
        parent_dir: PathLike = None
    ):
        if parent_dir:
            node = self.add_dir(parent_dir)
        else:
            node = self
        self.__from_structure(node, structure)

    def from_dir(
        self,
        search_dir: PathLike,
        recursive: bool = True
    ):
        self.root_path = Path(search_dir).resolve()
        directories = self.get_dirs(search_dir, recursive=recursive)
        for directory in directories:
            self.add_dir(directory)

    def add_dir(self, dir: PathLike):
        path = Path(dir).resolve()
        if not path.is_dir():
            raise InvalidError(path, 'directory')
        if not self.root_path:
            raise NotFoundError('root directory')
        if self.root_path == path:
            return self
        if not is_subpath(path, self.root_path):
            raise NotIncludeError(path, 'root directory')

        relpath = path.relative_to(self.root_path)
        directory_paths = [relpath, *list(relpath.parents)[:-1]]
        directory_paths.reverse()
        node = self
        for directory_path in directory_paths:
            node_path = "/"+str(directory_path.as_posix())
            directory_node = self.find(
                match=lambda node: node.path == node_path
            )
            if not directory_node:
                is_subspace = self.is_subspace(
                    Path(self.root_path, directory_path))
                directory = Directory(
                    path=directory_path,
                    is_subspace=is_subspace
                )
                directory_node = node.add(directory)
            node = directory_node
        return node

    @staticmethod
    def is_subspace(path: Path) -> bool:
        """Return true if path is a subspace entity .

        Args:
            path (Entity): The path to be checked.

        Returns:
            bool: Return ture if is valid entity.
        """
        path = path.resolve()
        is_subspace = Path(path, 'Subspace.yml').is_file()
        is_void = 'Void' in path.name.split("_")
        return is_subspace or is_void

    @classmethod
    def get_dirs(
        cls,
        search_dir: PathLike,
        recursive: bool = True
    ) -> list[Path]:
        search_path: Path = Path(search_dir).resolve()
        directories: list[Path] = []

        if recursive:
            # FIXME: replace this with Path.walk (python 3.12)
            # https://docs.python.org/3/library/pathlib.html
            for root, dirs, files in os.walk(search_path):
                for dir in dirs:
                    child = Path(root, dir).resolve()
                    directories.append(child)
        else:
            for child in search_path.iterdir():
                directories.append(child)
        return directories

    def make_dirs(self, root_dir: PathLike = None):
        if not self.root_path and not root_dir:
            raise NotFoundError('root directory')
        self.__make_dirs(
            path=Path(root_dir) or self.root_path,
            node=self
        )

    @classmethod
    def __make_dirs(
        cls,
        path: Path,
        node: Union[Tree, Node],
        is_subspace: bool = False
    ):
        if (isinstance(node, Node)):
            dirpath = Path(path, node.path.removeprefix("/")).resolve()
            is_subspace = node.data.is_subspace
        else:
            dirpath = path.resolve()

        dirpath.mkdir(parents=True, exist_ok=True)
        if is_subspace:
            with Path(dirpath, "Subspace.yml").open('w') as file:
                pass

        for sub_node in node:
            cls.__make_dirs(path, node=sub_node)

    def __from_structure(
        self,
        node: Union[Tree, Node],
        structure: Structure,
    ):
        for key in structure.keys():
            parent_path = node.path if isinstance(node, Node) else "."
            directory_name = format_name(key)
            directory_path = Path(parent_path, directory_name)
            node_path = "/"+str(directory_path.as_posix())
            directory_node = self.find(
                match=lambda node: node.path == node_path
            )
            if not directory_node:
                is_subspace = key[0] == "*" \
                    or "Void" in directory_name.split("_")
                directory = Directory(
                    path=directory_path,
                    is_subspace=is_subspace
                )
                directory_node = node.add(directory)
            sub_structure = structure[key]
            if sub_structure:
                self.__from_structure(
                    node=directory_node,
                    structure=sub_structure,
                )

    def render_tree(self):
        label_dict = {
            "📂": "ordinary direcotry\nwhich is for file classification.",
            "🎯": "subspace direcotry\nwhich is for presenting subspace.",
        }
        return Grid(
            self.format(repr="{node.data.ui_name}", title="(Root)"),
            Instruction(label_dict)
        )



import click
from omoospace.commands.common import choose_omoospace, to_omoospace
from omoospace.common import console


@click.command('to', help="Switch to another omoospace.")
@click.argument("omoospace", required=False, default='.',
                type=click.Path(exists=True, file_okay=False))
def cli_to(omoospace):
    try:
        omoospace = to_omoospace(omoospace)
    except Exception as err:
        console.print(err, style="warning")
        try:
            omoospace = choose_omoospace()
        except Exception as err:
            console.print(err, style="warning")



import click
from pathlib import Path
import validators
from InquirerPy import inquirer
from InquirerPy.validator import EmptyInputValidator
from InquirerPy.base import Choice
from prompt_toolkit.validation import ValidationError, Validator

from omoospace.directory import DirectoryTree
from omoospace.setting import Setting
from omoospace.subspace import Subspace, SubspaceTree
from omoospace.types import Creator, Software, Work
from omoospace.ui import Board, CreatorProfile, Info, Transfer
from omoospace.utils import find_first, format_name, is_subpath
from omoospace.omoospace import Omoospace
from omoospace.commands.common import collect_item_prompt, get_omoospace, option_omoospace, get_working_dir, inquirer_style, set_working_dir
from omoospace.common import console

SET_SUBSPACE_HELP = "Add or update subspace directory."
SET_PROCESS_HELP = "Add or update process directory."
SET_CREATOR_HELP = "Add or update creator."
SET_SOFTWARE_HELP = "Add or update software."
SET_WORK_HELP = "Add or update work filepath."


def _set_subspace(omoospace):
    omoospace: Omoospace = get_omoospace(omoospace)
    
    class SubspaceParentDirValidator(Validator):
        def validate(self, document):
            # Check if is empty
            if not len(document.text) > 0:
                raise ValidationError(
                    message="Parent directory cannot be empty.",
                    cursor_position=document.cursor_position,
                )

            # Check if is valid directory
            path = Path(document.text)
            if not path.is_dir():
                raise ValidationError(
                    message="'%s' is invalid directory." % document.text,
                    cursor_position=document.cursor_position,
                )

            # Check if is in SourceFiles
            if not is_subpath(path, omoospace.sourcefiles_path, or_equal=True):
                raise ValidationError(
                    message="'%s' is out of SourceFiles." % document.text,
                    cursor_position=document.cursor_position,
                )

    parent_dir = inquirer.filepath(
        message="Enter subspace parent directory",
        only_directories=True,
        validate=SubspaceParentDirValidator(),
        long_instruction='Press Tab to summon auto completer. \n'
        'Press Enter to confirm.',
        default=get_working_dir(),
        style=inquirer_style
    ).execute()
    set_working_dir(parent_dir)

    subspace_name = inquirer.text(
        message="Enter subspace name",
        validate=EmptyInputValidator("Name cannot be empty."),
        style=inquirer_style
    ).execute()
    subspace_pathname = format_name(subspace_name)

    subspace_comment = inquirer.text(
        message="Enter a comment of this subspace (Optional)",
        style=inquirer_style
    ).execute()

    subspace_info = {
        "name": subspace_name,
        "comments": subspace_comment or None
    }
    
    subspace_path = Path(parent_dir, subspace_pathname).resolve()
    subspace_tree = SubspaceTree()
    parent_node = subspace_tree.add_entity(parent_dir)
    parent_node.add(Subspace(pathname=subspace_pathname))
    # TODO: display new subspace with true name not dirname only.
    console.print(Board(
        Info("Name", "%s [dim](%s)[/dim]" %
             (subspace_name, subspace_path)),
        Info("Comments", subspace_comment),
        Info("Subspace Tree", subspace_tree.render_tree()),
        title="Set Subspace Form"
    ))

    is_confirm = inquirer.confirm(
        message="Confirm",
        default=True,
        style=inquirer_style
    ).execute()
    if is_confirm:
        omoospace.set_subspace(
            subspace=subspace_info,
            parent_dir=parent_dir,
        )


def _set_process(omoospace):
    omoospace: Omoospace = get_omoospace(omoospace)

    class ProcessParentDirValidator(Validator):
        def validate(self, document):
            # Check if is empty
            if not len(document.text) > 0:
                raise ValidationError(
                    message="Parent directory cannot be empty.",
                    cursor_position=document.cursor_position,
                )

            # Check if is valid directory
            path = Path(document.text)
            if not path.is_dir():
                raise ValidationError(
                    message="'%s' is invalid directory." % document.text,
                    cursor_position=document.cursor_position,
                )

            # Check if is in SourceFiles
            if not is_subpath(path, omoospace.sourcefiles_path, or_equal=True):
                raise ValidationError(
                    message="'%s' is out of SourceFiles." % document.text,
                    cursor_position=document.cursor_position,
                )

    parent_dir = inquirer.filepath(
        message="Enter process parent directory",
        only_directories=True,
        validate=ProcessParentDirValidator(),
        long_instruction='Press Tab to summon auto completer. \n'
        'Press Enter to confirm.',
        default=get_working_dir(),
        style=inquirer_style
    ).execute()
    set_working_dir(parent_dir)

    templates = Setting().process_templates
    template_choices = [
        Choice(name="*Custom", value=-1),
        *[Choice(
            name="%s (%s)" % (template.name, template.description),
            value=id
        ) for id, template in enumerate(templates)]
    ]

    template_id: int = inquirer.select(
        message="Choose a template",
        choices=template_choices,
        style=inquirer_style,
    ).execute()

    if template_id >= 0:
        while True:
            template = templates[template_id]
            console.print(template.tree.render_tree())
            confirm_id: int = inquirer.select(
                message="Choose the same again to confirm",
                choices=template_choices,
                default=template_id,
                long_instruction="Press Enter key to confirm",
                style=inquirer_style,
            ).execute()
            if confirm_id == template_id:
                break
            else:
                template_id = confirm_id
        process_structure = template.structure
    else:
        process_name = inquirer.text(
            message="Enter process name",
            validate=EmptyInputValidator("Name cannot be empty."),
            style=inquirer_style
        ).execute()
        process_structure = {format_name(process_name): None}

    directory_tree = DirectoryTree()
    directory_tree.root_path = omoospace.sourcefiles_path
    directory_tree.from_structure(process_structure, parent_dir)

    console.print(Board(
        Info("Directory Tree", directory_tree.render_tree()),
        title="Set Subspace Form"
    ))

    is_confirm = inquirer.confirm(
        message="Confirm",
        default=True,
        style=inquirer_style
    ).execute()
    if is_confirm:
        omoospace.set_process(
            process=process_structure,
            parent_dir=parent_dir,
        )


def _set_creator(omoospace):
    omoospace: Omoospace = get_omoospace(omoospace)

    class EmailValidator(Validator):

        def validate(self, document):
            # Allow empty
            if not validators.email(document.text):
                raise ValidationError(
                    message="Invaild email address.",
                    cursor_position=document.cursor_position,
                )

    class WebsiteValidator(Validator):

        def validate(self, document):
            # Allow empty
            if len(document.text) > 0 and not validators.url(document.text):
                raise ValidationError(
                    message="Invaild website address.",
                    cursor_position=document.cursor_position,
                )
    registered_creators = Setting().registered_creators
    creator_choices = [
        Choice(name="*Register", value=-1),
        *[Choice(
            name="%s (%s)" % (creator.get("name"), creator.get("email")),
            value=id
        ) for id, creator in enumerate(registered_creators)]
    ]

    creator_id = inquirer.fuzzy(
        message="Choose from registered creators",
        choices=creator_choices,
        style=inquirer_style
    ).execute()

    if creator_id >= 0:
        creator_info = registered_creators[creator_id]
    else:
        creator_email = inquirer.text(
            message="Enter creator email",
            validate=EmailValidator(),
            style=inquirer_style
        ).execute()
        creator_info = find_first(registered_creators, "email", creator_email)
        if not creator_info:
            creator_name = inquirer.text(
                message="Enter creator name",
                validate=EmptyInputValidator("Name cannot be empty."),
                style=inquirer_style
            ).execute()
            creator_website = inquirer.text(
                message="Enter creator website (Optional)",
                validate=WebsiteValidator(),
                long_instruction='Start with "http(s)://".',
                style=inquirer_style
            ).execute()
            creator_info: Creator = {
                "name": creator_name,
                "email": creator_email or None,
                "website": creator_website or None,
            }
            registered_creators.append(creator_info)
            Setting().registered_creators = registered_creators
        else:
            console.print(
                "Found creator with this email in regitered.", style="warning")

    console.print(CreatorProfile(creator_info))
    role_choices = Setting().role_choices
    creator_role = inquirer.fuzzy(
        message="Enter creator role in this omoospace (Optional)",
        choices=["*NotInclude", *role_choices],
        style=inquirer_style
    ).execute()
    if creator_role == "*NotInclude":
        creator_role = inquirer.text(
            message="Enter role name",
            validate=EmptyInputValidator("Name cannot be empty."),
            style=inquirer_style
        ).execute()

    creator_info["role"] = creator_role or None

    is_confirm = inquirer.confirm(
        message="Confirm",
        default=True,
        style=inquirer_style
    ).execute()
    if is_confirm:
        role_choices.append(creator_role)
        Setting().role_choices = role_choices
        omoospace.set_creator(creator_info)


def _set_software(omoospace):
    omoospace: Omoospace = get_omoospace(omoospace)
    software_choices = Setting().software_choices
    software_name = inquirer.fuzzy(
        message="Choose from presets",
        choices=["*NotInclude", *software_choices],
        style=inquirer_style
    ).execute()
    if software_name == "*NotInclude":
        software_name = inquirer.text(
            message="Enter software name",
            validate=EmptyInputValidator("Name cannot be empty."),
            style=inquirer_style
        ).execute()

    software_version = inquirer.text(
        message="Enter software version (Optional)",
        style=inquirer_style
    ).execute()

    software_info: Software = {
        "name": software_name,
        "version": software_version or None,
        "plugins": None,
    }
    console.print(software_info)
    is_confirm = inquirer.confirm(
        message="Confirm",
        default=True,
        style=inquirer_style
    ).execute()
    if is_confirm:
        software_choices.append(software_name)
        Setting().software_choices = software_choices
        omoospace.set_software(software_info)


def _set_work(omoospace):
    omoospace: Omoospace = get_omoospace(omoospace)
    work_paths: list[Path] = collect_item_prompt(omoospace.contents_path)

    work_name = inquirer.text(
        message="Enter work name (Optional)",
        style=inquirer_style
    ).execute()

    work_info = {
        "name": work_name,
        "paths": work_paths
    }
    console.print(work_paths)
    is_confirm = inquirer.confirm(
        message="Confirm",
        default=True,
        style=inquirer_style
    ).execute()
    if is_confirm:
        omoospace.set_work(work_info)


@click.group(
    'set',
    help="Add or update stuffs to omoospace.",
    invoke_without_command=True
)
@click.option(
    "-s", "--subspace",
    is_flag=True,
    help=SET_SUBSPACE_HELP
)
@click.option(
    "-p", "--process",
    is_flag=True,
    help=SET_PROCESS_HELP
)
@click.option(
    "-c", "--creator",
    is_flag=True,
    help=SET_CREATOR_HELP
)
@click.option(
    "-so", "--software",
    is_flag=True,
    help=SET_SOFTWARE_HELP
)
@click.option(
    "-w", "--work",
    is_flag=True,
    help=SET_WORK_HELP
)
@option_omoospace()
@click.pass_context
def cli_set(ctx, subspace, process, creator, software, work, omoospace):
    if not ctx.invoked_subcommand:
        if subspace:
            _set_subspace(omoospace)
        elif process:
            _set_process(omoospace)
        elif creator:
            _set_creator(omoospace)
        elif software:
            _set_software(omoospace)
        elif work:
            _set_work(omoospace)
        else:
            console.print("set something")


@click.command('subspace', help=SET_SUBSPACE_HELP)
@option_omoospace()
def cli_set_subspace(omoospace):
    _set_subspace(omoospace)


@click.command('process', help=SET_PROCESS_HELP)
@option_omoospace()
def cli_set_process(omoospace):
    _set_process(omoospace)


@click.command('creator', help=SET_CREATOR_HELP)
@option_omoospace()
def cli_set_creator(omoospace):
    _set_creator(omoospace)


@click.command('software', help=SET_SOFTWARE_HELP)
@option_omoospace()
def cli_set_software(omoospace):
    _set_software(omoospace)


@click.command('work', help=SET_WORK_HELP)
@option_omoospace()
def cli_set_work(omoospace):
    _set_work(omoospace)


cli_set.add_command(cli_set_subspace)
cli_set.add_command(cli_set_process)
cli_set.add_command(cli_set_creator)
cli_set.add_command(cli_set_software)
cli_set.add_command(cli_set_work)




from pathlib import Path
import shutil
from typing import Union
from zipfile import ZipFile

from rich.progress import track
from omoospace.directory import DirectoryTree
from omoospace.exceptions import CreateFailed, EmptyError, ExistsError, MoveFailed, NotFoundError, NotIncludeError
from omoospace.package import Package
from omoospace.subspace import SubspaceTree
from omoospace.common import console, yaml
from omoospace.ui import Board, Card, Info, ItemList, Table
from omoospace.utils import format_name, replace_or_append, reveal_in_explorer, is_subpath, PathLike, copy_to_path
from omoospace.types import Creator, Entity, Item, OmoospaceInfo, PackageInfo, Software, Structure, SubspaceInfo, Work

# TODO: package link in summary list

# TODO: annition to class and funtion


class OmoospaceTree(DirectoryTree):

    MAIN_DIRS = [
        "Contents",
        "ExternalData",
        "SourceFiles",
        "References",
        "StagedData"
    ]

    def __init__(self, structure: Structure = None) -> None:
        self.structure = structure or {}
        for dirname in self.MAIN_DIRS:
            if self.structure.get(dirname) is None:
                self.structure[dirname] = None
        super().__init__(structure=self.structure)


class Omoospace:
    """The class of omoospace.

    An omoospace class instance is always refer to a existed omoospace directory, not in ideal. 

    Attributes:
        name (str): Omoospace's name.
        description (str): Omoospace's description.
        creators (list[Creator]): Creator list.
        softwares (list[Software]): Software list.
        works (list[Work]): Work list.
        root_path (Path): Root path.
        sourcefiles_path (Path): SourceFiles directory path.
        contents_path (Path): Contents directory path.
        externaldata_path (Path): ExternalData directory path.
        stageddata_path (Path): StagedData directory path.
        references_path (Path): Reference directory path.
        info_path (Path): Omoospace.yml file path.
    """

    name: str
    description: str
    creators: list[Creator]
    softwares: list[Software]
    works: list[Work]

    def __init__(self, detect_dir: PathLike):
        """Initialize from existed omoospace directory.

        Args:
            detect_dir (PathLike): The start directory for detecting omoospace. It could be the subdirectories of omoospace.

        Raises:
            NotFoundError: No omoospace detected.
        """
        detect_path = Path(detect_dir).resolve()
        detect_path_parents = [detect_path, *detect_path.parents]
        omoospace_path = None
        for detect_path_parent in detect_path_parents:
            omoospace_info_path = Path(detect_path_parent, 'Omoospace.yml')
            if omoospace_info_path.exists():
                omoospace_path = detect_path_parent
                break
        if not omoospace_path:
            raise NotFoundError("omoospace", detect_dir)

        # path assignment
        self.root_path = omoospace_path
        self.sourcefiles_path = Path(self.root_path, 'SourceFiles').resolve()
        self.contents_path = Path(self.root_path, 'Contents').resolve()
        self.externaldata_path = Path(self.root_path, 'ExternalData').resolve()
        self.stageddata_path = Path(self.root_path, 'StagedData').resolve()
        self.references_path = Path(self.root_path, 'References').resolve()
        self.info_path = Path(self.root_path, "Omoospace.yml").resolve()

    def __read_info_file(self) -> OmoospaceInfo:
        with self.info_path.open('r', encoding='utf-8') as file:
            # aviod empty or invalid ifle
            info = yaml.load(file) or {}
        return info

    def __write_info_file(self, info: OmoospaceInfo):
        with self.info_path.open('w', encoding='utf-8') as file:
            yaml.dump(info, file)

    def __getattr__(self, name):
        if name in OmoospaceInfo.__annotations__.keys():
            info = self.__read_info_file()
            # FIXME: list attribute is not None but empty list.
            # may have better solution.
            if name[-1] == "s":
                return info.get(name) or []
            else:
                return info.get(name)

        return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        if name in OmoospaceInfo.__annotations__.keys():
            info = self.__read_info_file()
            info[name] = value
            self.__write_info_file(info)
        object.__setattr__(self, name, value)

    def is_omoospace_item(self, item: Item) -> bool:
        exists = item.exists()
        in_omoospace = is_subpath(item, self.root_path)
        not_in_stagedata = not is_subpath(
            item, self.stageddata_path, or_equal=True)
        not_omoospace_yml = item.name != "Omoospace.yml"
        return exists and in_omoospace and not_omoospace_yml and not_in_stagedata

    @property
    def subspace_entities(self) -> list[Entity]:
        """list[Entity]: the subspace entities in the sourcefiles directory."""
        return SubspaceTree.get_entities(search_dir=self.sourcefiles_path)

    @property
    def subspace_tree(self) -> SubspaceTree:
        """SubspaceTree: The subspace tree."""
        return SubspaceTree(search_dir=self.sourcefiles_path)

    @property
    def directory_tree(self,) -> DirectoryTree:
        """DirectoryTree: The directory tree."""
        return DirectoryTree(search_dir=self.root_path)

    @property
    def imported_packages(self) -> list[Package]:
        """list[Package]: The list of Package objects that are imported from ExternalData."""
        subdirs = [subdir for subdir in self.externaldata_path.iterdir()
                   if subdir.is_dir()]

        # Collect all Packages in ExternalData
        packages = []
        for subdir in subdirs:
            info_filepath = Path(subdir, 'Package.yml').resolve()
            if info_filepath.is_file():
                packages.append(Package(subdir))

        return packages

    def set_subspace(
            self,
            subspace: Union[SubspaceInfo, str],
            parent_dir: PathLike = None,
            reveal_when_success: bool = True,
            collect_entities: bool = True
    ) -> Path:
        """Add or update subspace to this omoospace.

        Args:
            subspace (Union[SubspaceInfo, str]): [description]
            parent_dir (PathLike, optional): [description]. Defaults to None.
            reveal_when_success (bool, optional): [description]. Defaults to True.
            collect_entities (bool, optional): [description]. Defaults to True.

        Raises:
            ExistsError: [description]
            NotIncludeError: [description]
            EmptyError: [description]
            EmptyError: [description]
            EmptyError: [description]
            MoveFailed: [description]

        Returns:
            Path: [description]
        """

        parent_path = Path(parent_dir).resolve(
        ) if parent_dir else self.sourcefiles_path

        # Check if is valid directory
        if not parent_path.is_dir():
            raise ExistsError(parent_dir)

        # Check if is in SourceFiles
        if not is_subpath(parent_path, self.sourcefiles_path, or_equal=True):
            raise NotIncludeError(parent_dir, "SourceFiles")
        # FIXME: using SubspaceInfo need more code base on typing.
        # do it later.
        if isinstance(subspace, dict):
            # Check if name is empty
            if not subspace.get("name"):
                raise EmptyError("name")
            if not len(subspace.get("name")) > 0:
                raise EmptyError("name")
            subspace_info = subspace
            subspace_pathname = format_name(subspace.get("name"))
        else:
            if not len(subspace) > 0:
                raise EmptyError("name")

            subspace_info = {
                "name": subspace,
                "comments": None
            }
            subspace_pathname = format_name(subspace)

        subspace_path = Path(parent_path, subspace_pathname).resolve()

        # make subspace dir
        subspace_path.mkdir(parents=True, exist_ok=True)
        subspace_info_path = Path(subspace_path, 'Subspace.yml')
        with subspace_info_path.open('w', encoding='utf-8') as file:
            yaml.dump(subspace_info, file)

        if (collect_entities):
            try:
                entities = SubspaceTree.get_entities(
                    parent_path, recursive=False)

                def is_match(entity: Entity):
                    entity_stem = entity.stem
                    not_itself = entity_stem != subspace_pathname

                    is_match = False
                    entity_namespaces = format_name(entity_stem).split('_')
                    subspace_namespaces = subspace_pathname.split('_')
                    for i in range(len(subspace_namespaces)):
                        subspace_suffix = '_'.join(subspace_namespaces[i:])
                        entity_prefix = '_'.join(
                            entity_namespaces[:len(subspace_namespaces) - i])
                        if subspace_suffix == entity_prefix:
                            is_match = True
                            break
                    return not_itself and is_match

                # Remove entity that not match the name
                entities = list(filter(is_match, entities))

                for entity in entities:
                    shutil.move(entity, Path(
                        subspace_path, entity.name).resolve())
            except:
                raise MoveFailed('entities')

        if reveal_when_success:
            reveal_in_explorer(subspace_path)

        return subspace_path

    def set_process(
            self,
            process: Union[Structure, list[str], str],
            parent_dir: PathLike = None,
            reveal_when_success: bool = True
    ) -> Path:
        """Add or update process to this omoospace.

        Args:
            process (Union[Structure, list[str], str]): [description]
            parent_dir (PathLike, optional): [description]. Defaults to None.
            reveal_when_success (bool, optional): [description]. Defaults to True.

        Raises:
            ExistsError: [description]
            NotIncludeError: [description]
            CreateFailed: [description]

        Returns:
            Path: [description]
        """
        parent_path = Path(parent_dir).resolve(
        ) if parent_dir else self.sourcefiles_path

        # Check if is valid directory
        if not parent_path.is_dir():
            raise ExistsError(parent_dir)

        # Check if is in SourceFiles
        if not is_subpath(parent_path, self.sourcefiles_path, or_equal=True):
            raise NotIncludeError(parent_dir, "SourceFiles")

        if isinstance(process, str):
            process = [process]

        if isinstance(process, list):
            process_structure = {}
            for p in process:
                process_structure[p] = None
        else:
            process_structure = process

        try:
            DirectoryTree(process_structure).make_dirs(parent_path)
        except Exception as err:
            raise CreateFailed("process directories")

        if reveal_when_success:
            reveal_in_explorer(parent_path)

    def set_creator(self, creator: Creator):
        """Add or update creator to this omoospace.

        Args:
            creator (Creator): [description]
        """
        creators = self.creators or []
        self.creators = replace_or_append(creators, creator, 'email')

    def set_software(self, software: Software):
        """Add or update software to this omoospace.

        Args:
            software (Software): [description]
        """
        softwares = self.softwares or []
        self.softwares = replace_or_append(softwares, software, 'name')

    def set_work(self, work: Work):
        """Add or update work to this omoospace.

        Args:
            work (Work): [description]

        Raises:
            NotFoundError: [description]
            NotFoundError: [description]
        """
        if "paths" not in work.keys():
            raise NotFoundError("paths", work)

        paths = [Path(path).resolve() for path in work["paths"]]
        paths = [path for path in paths
                 if path.exists() and is_subpath(path, self.contents_path)]

        if len(paths) == 0:
            raise NotFoundError("valid paths", work)

        relpaths = [path.relative_to(self.contents_path) for path in paths]

        work = {
            "name": work.get("name") or relpaths[0].stem,
            "paths": [str(relpath.as_posix())for relpath in relpaths]
        }
        works = self.works or []
        self.works = replace_or_append(works, work, 'name')

    def export_package(
        self,
        items: list[PathLike],
        name: str = None,
        export_dir: PathLike = None,
        info: PackageInfo = None,
        reveal_when_success: bool = True,
        overwrite_existing: bool = True
    ) -> Package:
        """Export a package.

        Args:
            items (list[PathLike]): [description]
            name (str, optional): [description]. Defaults to None.
            export_dir (PathLike, optional): [description]. Defaults to None.
            info (PackageInfo, optional): [description]. Defaults to None.
            reveal_when_success (bool, optional): [description]. Defaults to True.
            overwrite_existing (bool, optional): [description]. Defaults to True.

        Raises:
            ExistsError: [description]
            Exception: [description]
            Exception: [description]

        Returns:
            Package: [description]
        """
        name = name or self.name
        export_path = Path(export_dir).resolve() if export_dir \
            else Path(self.root_path, 'StagedData', 'Packages').resolve()
        package_dirname = format_name(name)
        package_path = Path(export_path, package_dirname).resolve()

        # Check if package dir exists
        if (package_path.is_dir()):
            if (overwrite_existing):
                shutil.rmtree(package_path, ignore_errors=True)
            else:
                raise ExistsError('package', package_path)

        items: list[Path] = [Path(item).resolve() for item in items]
        items = list(filter(self.is_omoospace_item, items))

        # Check if is enouph items
        if (len(items) == 0):
            raise Exception('Export Failed, at least one item')

        info = info or {}
        package_info = {
            "name": info.get("name") or name,
            "version": info.get("version"),
            "description": info.get("description"),
            "creators": self.creators,
        }

        try:
            package_path.mkdir(parents=True, exist_ok=True)
            package_info_path = Path(package_path, 'Package.yml')
            package_md_path = Path(package_path, 'README.md')

            with package_info_path.open('w', encoding='utf-8') as file:
                yaml.dump(package_info, file)

            with package_md_path.open('w', encoding='utf-8') as file:
                file.write('# %s\n' % package_info.get("name"))
                file.write("%s\n" % package_info.get("description") or "")
                file.write('## Package Info\n')
                file.write('**version:** %s  \n' %
                           package_info.get("version") or "")
                file.write('**creators:**\n')
                for author in package_info.get("creators") or []:
                    file.write("- %s\n" % author.get("name") or "Unknow")
                file.write('## Items List\n')
                for item in items:
                    file.write("- %s\n" % item)

            for i in track(range(len(items)), description="Processing..."):
                item = items[i]
                item_relpath = item.relative_to(self.root_path)
                copy_to_path(item, Path(package_path, item_relpath))

        except Exception as err:
            # Delete all if failed.
            shutil.rmtree(package_path, ignore_errors=True)
            raise Exception("Fail to export Package", err)

        if reveal_when_success:
            reveal_in_explorer(package_path)

        return Package(package_path)

    def import_package(
        self,
        import_dir: PathLike,
        reveal_when_success: bool = True,
        overwrite_existing: bool = True
    ):
        """Imports the package into the ExternalData directory.

        Args:
            import_dir (PathLike): [description]
            reveal_when_success (bool, optional): [description]. Defaults to True.
            overwrite_existing (bool, optional): [description]. Defaults to True.

        Raises:
            ExistsError: [description]
        """
        # get package form import directory
        import_path = Path(import_dir).resolve()
        package = Package(import_path)

        # check if destination directory exists
        package_path = Path(
            self.root_path, 'ExternalData', format_name(package.name))
        if (package_path.is_dir()):
            if (overwrite_existing):
                shutil.rmtree(package_path, ignore_errors=True)
            else:
                raise ExistsError('package', package_path)

        if (import_path.suffix == ".zip"):
            with ZipFile(import_path, 'r') as zip:
                zip.extractall(package_path)
        else:
            copy_to_path(import_path, package_path)

        if reveal_when_success:
            reveal_in_explorer(package_path)

    def show_summary(
        self,
        output_dir: PathLike = None,
        reveal_when_success: bool = True,
    ):
        """Print a summary of the omoospace.

        Args:
            output_dir (PathLike, optional): [description]. Defaults to None.
        """
        imported_packages = self.imported_packages
        subspace_tree = self.subspace_tree

        ui_creators = ItemList([
            "%s [dim]%s[/dim]" % (creator.get("name"), creator.get("role"))
            for creator in self.creators])
        ui_softwares = ItemList([
            "%s [dim]%s[/dim]" % (software.get("name"),
                                  software.get("version"))
            for software in self.softwares])

        ui_works = Table(
            "Name",
            "Paths"
        )

        for work in self.works:
            ui_works.add_row(
                work.get("name"),
                "\n".join(work.get("paths") or [])
            )

        console.print(Board(
            Info("Name", "%s [dim](%s)[/dim]" %
                 (self.name, self.root_path)),
            Info("Description", self.description),
            Info("Creators", ui_creators),
            Info("Softwares", ui_softwares),
            Info("Works", ui_works),
            Info("Subspace Tree", subspace_tree.render_tree()),
            Info("Subspace Entities", subspace_tree.render_table()),
            title="Summary"
        ))

        if (output_dir):
            subspace_tree.draw_graph(output_dir, reveal_when_success)

    def show_info(self):
        """Show profile of this omoospace.
        """
        ui_creators = ItemList([
            "%s [dim]%s[/dim]" % (creator.get("name"), creator.get("role"))
            for creator in self.creators])
        ui_softwares = ItemList([
            "%s [dim]%s[/dim]" % (software.get("name"),
                                  software.get("version"))
            for software in self.softwares])
        ui_works = Table(
            "Name",
            "Paths"
        )
        # TODO: add [link] to path
        for work in self.works:
            ui_works.add_row(
                work.get("name"),
                "\n".join(work.get("paths") or [])
            )
        console.print(Board(
            Info("Name", "%s [dim](%s)[/dim]" %
                 (self.name, self.root_path)),
            Info("Description", self.description),
            Info("Creators", ui_creators),
            Info("Softwares", ui_softwares),
            Info("Works", ui_works),
            title="Info"
        ))

    def show_subspace_tree(self):
        """Shows the subspace tree.
        """
        subspace_tree = self.subspace_tree
        console.print(Board(
            subspace_tree.render_tree(),
            title="Subspace Tree"
        ))

    def show_directory_tree(self):
        """Shows the directory tree.
        """
        directory_tree = self.directory_tree
        console.print(Board(
            directory_tree.render_tree(),
            title="Directory Tree"
        ))

    def show_imported_packages(self):
        """Show imported packages.
        """
        imported_packages = self.imported_packages
        ui_imported_packages = Table(
            "Directory", "Description",
            rows=[[package.name, package.description]
                  for package in imported_packages]
        )
        console.print(Board(
            ui_imported_packages,
            title="Imported Packages"
        ))

    def show_subspace_entities(self):
        """Shows the subspace entities.
        """
        subspace_tree = self.subspace_tree
        console.print(Board(
            subspace_tree.render_table(),
            title="Subspace Entities"
        ))

    @classmethod
    def create(
        cls,
        name: str,
        create_dir: PathLike = '.',
        info: OmoospaceInfo = None,
        structure: Structure = None,
        reveal_when_success: bool = True
    ):
        """Create a Omoospace .

        Args:
            name (str): [description]
            create_dir (PathLike, optional): [description]. Defaults to '.'.
            info (OmoospaceInfo, optional): [description]. Defaults to None.
            structure (OmoospaceStructure, optional): [description]. Defaults to None.
            reveal_when_success (bool, optional): [description]. Defaults to True.

        Raises:
            ExistsError: [description]
            Exception: [description]

        Returns:
            [type]: [description]
        """
        omoospace_dirname = format_name(name)
        omoospace_path = Path(create_dir, omoospace_dirname).resolve()

        # Check if omoospace exists
        if (omoospace_path.is_dir()):
            raise ExistsError('omoospace', create_dir)

        # write omoospace info to yml
        info = info or {}
        omoospace_info: OmoospaceInfo = {
            "name": info.get("name") or name,  # can be any name style
            "description": info.get("description"),
            "creators": info.get("creators"),
            "softwares": info.get("softwares"),
            "works": info.get("works"),
        }

        try:
            # create dirs
            OmoospaceTree(structure).make_dirs(omoospace_path)
            omoospace_info_path = Path(
                omoospace_path, 'Omoospace.yml').resolve()
            with omoospace_info_path.open('w', encoding='utf-8') as file:
                yaml.dump(omoospace_info, file)
        except Exception as err:
            shutil.rmtree(omoospace_path, ignore_errors=True)
            raise CreateFailed("omoospace directories")

        if reveal_when_success:
            reveal_in_explorer(omoospace_path)

        return cls(omoospace_path)



import click
from pathlib import Path
from InquirerPy import inquirer, get_style
from prompt_toolkit.validation import ValidationError, Validator

from omoospace.ui import Transfer
from omoospace.exceptions import InvalidError, NotFoundError
from omoospace.setting import Setting
from omoospace.types import PathLike
from omoospace.utils import is_subpath
from omoospace.common import console
from omoospace.omoospace import Omoospace


inquirer_style = get_style(
    {
        "questionmark": "#ffd700",
        "answermark": "#ffd700",
        "question": "",
        "answered_question": "",
        "pointer": "italic #ffd700",
        "input": "italic #ffd700",
        "answer": "italic #ffd700",
        "skipped": "#808080",
        "validator": "bold fg:#ffffff bg:red",
    },
    style_override=False
)


def option_omoospace():
    return click.option(
        "-os", "--omoospace",
        type=click.Path(),
        help="Set the working omoospace path."
    )


def choose_omoospace():
    recent_list = Setting().recent_omoospaces or []

    def is_valid(omoospace_dir):
        try:
            Omoospace(omoospace_dir)
            return True
        except:
            return False

    recent_list = list(filter(is_valid, recent_list))
    Setting().recent_omoospaces = recent_list
    recent_list.reverse()

    if (len(recent_list) > 0):
        omoospace_dir = inquirer.select(
            message="Choose one from recent",
            choices=recent_list,
            style=inquirer_style
        ).execute()
        return to_omoospace(omoospace_dir)
    else:
        raise NotFoundError("recent omoospace", "setting")


def to_omoospace(detect_path: str):
    omoospace = Omoospace(detect_path)
    recent_list = Setting().recent_omoospaces or []
    omoospace_dir = str(omoospace.root_path)
    # remove old record in settings
    try:
        recent_list.remove(omoospace_dir)
    except:
        pass
    recent_list.append(omoospace_dir)
    Setting().recent_omoospaces = recent_list
    Setting().working_omoospace = omoospace_dir
    console.print("Current working at 🛠️ [bright]%s (%s)[/bright]." %
                  (omoospace.name, omoospace_dir))
    return omoospace


def get_omoospace(detect_path: str = None):
    detect_path = detect_path or "."
    try:
        omoospace = to_omoospace(detect_path)
    except:
        try:
            omoospace_dir = Setting().working_omoospace
            omoospace = to_omoospace(omoospace_dir)
        except Exception as err:
            console.print(err, style="warning")
            Setting().working_omoospace = None
            omoospace = choose_omoospace()
    return omoospace


def is_valid_working_dir(working_dir: PathLike):
    root_working_dir = get_root_working_dir()
    working_dir = working_dir or "."
    working_path = Path(working_dir).resolve()

    exists = working_path.is_dir()
    is_include = is_subpath(working_dir, root_working_dir, or_equal=True)
    return exists and is_include


def get_root_working_dir():
    omoospace_dir = Setting().working_omoospace
    if not omoospace_dir:
        raise NotFoundError("working omoospace", "setting")

    root_working_dir = str(Path(omoospace_dir, "SourceFiles").resolve())
    return root_working_dir


def get_working_dir():
    working_dir = Setting().working_directory
    if not is_valid_working_dir(working_dir):
        working_dir = get_root_working_dir()

    Setting().working_directory = working_dir
    return working_dir


def set_working_dir(working_dir: PathLike):
    working_dir = str(Path(working_dir).resolve())

    if not is_valid_working_dir(working_dir):
        raise InvalidError(working_dir, "working directory")

    Setting().working_directory = working_dir


def collect_item_prompt(root_dir: PathLike, item_filter=None):
    root_path = Path(root_dir).resolve()

    class ItemValidator(Validator):
        def validate(self, document):
            # Check if is empty
            pending_items = get_items(document.text)
            if len(pending_items) == 0:
                raise ValidationError(
                    message="Find no files or directories in %s" % document.text,
                    cursor_position=document.cursor_position,
                )

    def get_items(path_pattern: str):
        if not path_pattern:
            # inquirer will trigger filter even if is escaped. but path_pattern become None
            return []
        # FIXME: glob must be relative path, but inquirer need abs path for auto complete.
        # maybe can make a fake path auto completer to fix this

        def remove_prefix(string, prefix):
            for i in range(len(prefix)):
                end_index = len(prefix) - i
                sub_prefix = prefix[:end_index]
                new_string = string.removeprefix(sub_prefix)
                if new_string != string:
                    return new_string
            return string
        path_pattern = remove_prefix(path_pattern, str(root_path))
        path_pattern = path_pattern.removeprefix("\\")
        path_pattern = path_pattern.removeprefix("/")
        if path_pattern == "":
            return []
        items = sorted(root_path.glob(path_pattern))
        items = [item.resolve() for item in items]
        if item_filter:
            items = list(filter(item_filter, items))
        return items

    def print_items(pending, checked):
        pending = [item.relative_to(root_path)
                   for item in pending]
        checked = [item.relative_to(root_path)
                   for item in checked]
        transfer = Transfer(
            pending, checked, left_header="Pending", right_header="Checked")
        console.print(transfer)

    pending_items = []
    checked_items = []
    while True:
        if len(checked_items) > 0:
            print_items(pending_items, checked_items)
        pending_items = inquirer.filepath(
            message="Enter the item path or path pattern",
            long_instruction='Press Tab to summon auto completer.\n'
            'Press Enter to confirm the input.\n'
            'Press Esc to finish adding item (at least one checked item).',
            filter=get_items,
            validate=ItemValidator(),
            default=str(root_path),
            keybindings={"skip": [{"key": "escape"}]},
            mandatory=len(checked_items) == 0,
            style=inquirer_style
        ).execute()
        if len(pending_items) == 0:
            break

        print_items(pending_items, checked_items)

        check_in = inquirer.confirm(
            message="Check if the items on left are wanted",
            default=True,
            style=inquirer_style
        ).execute()

        if check_in:
            checked_items.extend(pending_items)
            checked_items = list(set(checked_items))
        pending_items = []

    return checked_items



import re
import os
from pathlib import Path
import shutil
from pypinyin import lazy_pinyin
from omoospace.exceptions import NotFoundError
from omoospace.types import PathLike
from omoospace.common import console
from omoospace.ui import Table

def format_name(string: str, verbose=False):
    debug_info = Table(
        {"header": "Step", "justify": "right"},
        {"header": "Result", "justify": "left", "style": "highlight"}
    )
    debug_info.add_row("Original", string)

    # [A-Za-z0-9_-] only
    string = re.sub(r'[^\w-]', ' ', string)
    parts = [string for string in string.split()]
    debug_info.add_row("to Parts",
                       " [white]|[/white] ".join(parts))

    # Chinese character to pinyin
    parts = [char
             for string in parts
             for char in lazy_pinyin(string)]

    debug_info.add_row("to PinYin",
                       " [white]|[/white] ".join(parts))

    # to PascalCase
    parts = [string.title() if string.islower() else string
             for string in parts]
    debug_info.add_row("to PascalCase",
                       " [white]|[/white] ".join(parts))
    string = ''.join(parts)
    if (verbose):
        console.print(debug_info)
    return string


def replace_or_append(
    object_list: list[dict],
    new_object: dict,
    key: str
):
    if key not in new_object.keys():
        raise NotFoundError(key, new_object)
    index = find_first_index(object_list, key, new_object.get(key))
    if index >= 0:
        object_list[index] = new_object
    else:
        object_list.append(new_object)
    return object_list


def find_first_index(
    object_list: list[dict],
    key: str,
    value: any
):
    index = -1
    for i, dict in enumerate(object_list):
        if dict.get(key) == value:
            index = i
            break

    return index


def find_first(
    object_list: list[dict],
    key: str,
    value: any
):
    index = find_first_index(object_list, key, value)

    if index >= 0:
        return object_list[index]
    else:
        return None


def reveal_in_explorer(dir: str):
    try:
        os.startfile(Path(dir).resolve())
    except Exception as err:
        print("Fail to reveal", err)


def is_subpath(child: PathLike, parent: PathLike, or_equal=False):
    """Return True if child is a subpath of parent .

    Args:
        child (str): [description]
        parent (str): [description]
        or_equal (bool, optional): [description]. Defaults to False.

    Returns:
        [type]: [description]
    """
    parent = Path(parent).resolve()
    child = Path(child).resolve()
    is_subpath = parent in child.parents
    is_equal = parent == child
    if or_equal:
        return is_subpath or is_equal
    else:
        return is_subpath


def copy_to_path(src: PathLike, dst: PathLike):
    """Copies the contents form src to dst .

    Args:
        src (Path): [description]
        dst (Path): [description]
    """
    src = Path(src).resolve()
    dst = Path(dst).resolve()
    dst.parent.mkdir(parents=True, exist_ok=True)
    if src.is_dir():
        shutil.copytree(src, dst)
    else:
        shutil.copy(src, dst)



import click
from InquirerPy import inquirer
from prompt_toolkit.validation import ValidationError, Validator

from omoospace.package import Package
from omoospace.commands.common import option_omoospace, get_omoospace, inquirer_style
from omoospace.ui import Board, Info
from omoospace.common import console


@click.command('import', help="Import a package.")
@option_omoospace()
def cli_import(omoospace):
    omoospace = get_omoospace(omoospace)

    class ImportPackageValidator(Validator):
        def validate(self, document):
            # Check if is empty
            if not len(document.text) > 0:
                raise ValidationError(
                    message="Package path cannot be empty.",
                    cursor_position=document.cursor_position,
                )

            # Check if is package
            try:
                package = Package(document.text)
            except:
                raise ValidationError(
                    message="Package path is not a valid package.",
                    cursor_position=document.cursor_position,
                )

    import_dir = inquirer.filepath(
        message="Enter package file|dir path",
        validate=ImportPackageValidator(),
        long_instruction='Press Tab to summon auto completer. \n'
        'Press Enter to confirm.',
        style=inquirer_style
    ).execute()

    package = Package(import_dir)
    console.print(Board(
        Info("Name", "%s [dim](%s)[/dim]" %
             (package.name, import_dir)),
        Info("Description", package.description),
        Info("Version", package.version),
        Info("Items", package.items),
        title="Import Package Form"
    ))

    is_confirm = inquirer.confirm(
        message="Confirm",
        default=True,
        style=inquirer_style
    ).execute()
    if is_confirm:
        omoospace.import_package(import_dir)
    console.print("Successfully import! 📦", style="cheer")



import click
from omoospace.commands.common import get_omoospace, option_omoospace

SHOW_INFO_HELP = "Show omoospace info."
SHOW_DIRECTORY_HELP = "Show omoospace directory tree."
SHOW_SUBSPACES_HELP = "Show omoospace subspace tree."
SHOW_ENTITIES_HELP = "Show omoospace entities."
SHOW_PACKAGES_HELP = "Show omoospace packages."
STRUCTURE_DIR_HELP = "Set Directory of structure.html for structure visualization"


def _show_summary(omoospace,output_dir):
    omoospace = get_omoospace(omoospace)
    omoospace.show_summary(output_dir)


def _show_info(omoospace):
    omoospace = get_omoospace(omoospace)
    omoospace.show_info()


def _show_directories(omoospace):
    omoospace = get_omoospace(omoospace)
    omoospace.show_directory_tree()


def _show_subspaces(omoospace):
    omoospace = get_omoospace(omoospace)
    omoospace.show_subspace_tree()


def _show_entities(omoospace):
    omoospace = get_omoospace(omoospace)
    omoospace.show_subspace_entities()


def _show_packages(omoospace):
    omoospace = get_omoospace(omoospace)
    omoospace.show_imported_packages()


@click.group(
    'show',
    help="Show omoospace statistics",
    invoke_without_command=True
)
@click.option(
    "-i", "--info",
    is_flag=True,
    help=SHOW_INFO_HELP
)
@click.option(
    "-d", "--directories",
    is_flag=True,
    help=SHOW_DIRECTORY_HELP
)
@click.option(
    "-s", "--subspaces",
    is_flag=True,
    help=SHOW_SUBSPACES_HELP
)
@click.option(
    "-e", "--entities",
    is_flag=True,
    help=SHOW_ENTITIES_HELP
)
@click.option(
    "-p", "--packages",
    is_flag=True,
    help=SHOW_PACKAGES_HELP
)
@option_omoospace()
@click.option(
    "-r", "--report-dir",
    type=click.Path(),
    help=STRUCTURE_DIR_HELP
)
@click.pass_context
def cli_show(ctx, info, directories, subspaces, entities, packages, omoospace, report_dir):
    if not ctx.invoked_subcommand:
        if info:
            _show_info(omoospace)
        elif directories:
            _show_directories(omoospace)
        elif subspaces:
            _show_subspaces(omoospace)
        elif entities:
            _show_entities(omoospace)
        elif packages:
            _show_packages(omoospace)
        else:
            _show_summary(omoospace, report_dir)


@click.command('info', help=SHOW_INFO_HELP)
@option_omoospace()
def cli_show_info(omoospace):
    _show_info(omoospace)


@click.command('directories', help=SHOW_DIRECTORY_HELP)
@option_omoospace()
def cli_show_directories(omoospace):
    _show_directories(omoospace)


@click.command('subspaces', help=SHOW_SUBSPACES_HELP)
@option_omoospace()
def cli_show_subspaces(omoospace):
    _show_subspaces(omoospace)


@click.command('entities', help=SHOW_PACKAGES_HELP)
@option_omoospace()
def cli_show_entities(omoospace):
    _show_entities(omoospace)


@click.command('packages', help=SHOW_ENTITIES_HELP)
@option_omoospace()
def cli_show_packages(omoospace):
    _show_packages(omoospace)


cli_show.add_command(cli_show_directories)
cli_show.add_command(cli_show_subspaces)
cli_show.add_command(cli_show_entities)
cli_show.add_command(cli_show_packages)




import os
from pathlib import Path
from typing import Union
from nutree import Tree, Node
from omoospace.exceptions import InvalidError

from omoospace.graph import draw_graph
from omoospace.types import Entity, PathLike, Route, SubspaceInfo, SubspaceType
from omoospace.common import console, yaml
from omoospace.ui import Grid, Instruction, Table
from omoospace.utils import format_name


class Subspace():
    def __init__(
        self,
        pathname: str,
        info: SubspaceInfo = None,
    ):
        self.pathname = pathname
        self.entities: list[Entity] = []
        self.flow: int = 0
        self.type: SubspaceType = None

        if isinstance(info, str):
            self.name = pathname
            self.comments = [info]
        elif isinstance(info, list):
            self.name = pathname
            self.comments = info
        elif isinstance(info, dict):
            self.name = info.get('name') or pathname
            comments = info.get('comments')
            if isinstance(comments, str) and comments:
                self.comments = [comments]
            elif isinstance(comments, list):
                self.comments = [*comments]
            else:
                self.comments = []
        else:
            self.name = pathname
            self.comments = []

    def __repr__(self):
        return self.pathname

    @property
    def ui_name(self):
        if self.type == SubspaceType.DIRECTORY:
            icon = "📁"
            link = self.entities[0]
            name_str = "[link=%s]%s[/link]" % (link, self.name)
        elif self.type == SubspaceType.FILE:
            icon = "📄"
            link = self.entities[0].parent
            name_str = "[link=%s]%s[/link]" % (link, self.name)
        elif self.type == SubspaceType.PHANTOM:
            icon = "💿"
            name_str = self.name
        else:
            icon = "⛔"
            name_str = self.name

        if self.pathname == self.name:
            pathname_str = ""
        else:
            pathname_str = " [dim](%s)[/dim]" % self.pathname
        return "%s %s%s" % (icon, name_str, pathname_str)

    @property
    def html_name(self):
        if self.pathname == self.name:
            pathname_str = ""
        else:
            pathname_str = " (%s)" % self.pathname
        return "%s%s" % (self.name, pathname_str)

    @property
    def ui_comments(self):
        return ["[dim]- %s[/dim]" %
                comment for comment in self.comments]

    @property
    def ui_entities(self):
        entity_links: list[str] = []
        for entity in self.entities:
            entity_path = entity if entity.is_dir() else entity.parent
            entity_links.append("[link=%s]%s[/link]" %
                                (entity_path, entity.name))
        return entity_links

    @property
    def html_entities(self):
        entity_links: list[str] = []
        for entity in self.entities:
            entity_links.append("&nbsp<a href='%s' target='_blank'>%s</a>&nbsp"
                                % (entity, entity.name))
        return entity_links


class SubspaceTree(Tree):
    """The class of subspace tree.

    An subspace tree instance is always refer to a existed SourceFiles directory, not in ideal. 

    """

    def __init__(
        self,
        search_dir: PathLike = None
    ):
        """Initialize the subspace tree from the given directory.

        Args:
            search_dir (PathLike): [description]
        """
        super().__init__()
        if (search_dir):
            self.from_dir(search_dir)

    @classmethod
    def get_entity_route(cls, entity_path: PathLike) -> Route:
        """Returns a route to a entity path.

        Args:
            entity_path (PathLike): [description]

        Returns:
            Route: The list of pathname of subspace.
        """
        subspace_tree = cls()
        entity_end_node = subspace_tree.add_entity(entity_path)
        return entity_end_node.path.split("/")[1:]

    def from_dir(
        self,
        search_dir: PathLike,
        recursive: bool = True
    ):
        """Set the subspace tree from the given directory.

        Args:
            search_dir (PathLike): [description]
            recursive (bool, optional): [description]. Defaults to True.
        """
        entities = self.get_entities(search_dir, recursive=recursive)
        for entity in entities:
            self.add_entity(entity)

    def add_entity(self, entity_path: PathLike) -> Node:
        """Add subspace node to the tree based on entity.

        The entity path must be real, not in ideal.

        Args:
            entity (Entity): [description]

        Returns:
            Route: List of subspace name along the route.

        Raises:
            InvalidError: entity path is invalid.
        """
        entity = Path(entity_path).resolve()
        if not self.is_entity(entity) == 0:
            InvalidError(entity, 'entity')

        # Get relpath parts. remove those directory that is not entity.
        entities = [entity]
        for parent in entity.parents:
            if self.is_entity(parent):
                entities.append(parent)
        entities.reverse()

        parts = [{
            "name": format_name(entity.stem),
            "entity": entity
        } for entity in entities]

        # Get route nodes. remove duplicate prefix.
        route: Route = []
        node = self
        for part in parts:
            part_name: str = part["name"]
            part_entity: Entity = part["entity"]

            namespaces: list[str] = part_name.split('_')
            subspace_info_path = Path(part_entity, "Subspace.yml").resolve()

            # clip matched prefix spaces
            for i in range(len(route)):
                route_str_suffix = '_'.join(route[i:])
                name_prefix = '_'.join(namespaces[:len(route) - i])
                if route_str_suffix == name_prefix:
                    namespaces = namespaces[len(route) - i:]
                    break

            # create or get subspace by namespace
            for i in range(len(namespaces)+1):
                # if all namespace is match,the entity will be add to current route
                # so the index is start from -1
                i = i - 1
                is_last = i == len(namespaces)-1
                subspace_route = route + namespaces[:i+1]
                if (len(subspace_route) == 0):
                    continue

                node_path = "/" + "/".join(subspace_route)
                subspace_node = self.find(
                    match=lambda node: node.path == node_path
                )
                if not subspace_node:
                    if is_last and subspace_info_path.is_file():
                        with subspace_info_path.open('r', encoding='utf-8') as file:
                            subspace_info = yaml.load(file) or {}
                    else:
                        subspace_info = {}
                    subspace = Subspace(
                        pathname=namespaces[i],
                        info=subspace_info,
                    )
                    subspace_node = node.add(subspace)

                # if is last, add enitity to it.
                if is_last and part_entity not in subspace_node.data.entities:
                    subspace_node.data.entities.append(part_entity)

                # assign type
                subspace_type = subspace_node.data.type
                if is_last and part_entity.is_dir():
                    subspace_node.data.type = SubspaceType.DIRECTORY
                elif subspace_type != SubspaceType.DIRECTORY \
                        and is_last and part_entity.is_file():
                    subspace_node.data.type = SubspaceType.FILE
                elif subspace_type != SubspaceType.DIRECTORY \
                        and subspace_type != SubspaceType.FILE\
                        and part_entity.exists():
                    subspace_node.data.type = SubspaceType.PHANTOM

                # what ever is end node or not, all nodes are passed through.
                subspace_node.data.flow += 1

                node = subspace_node
            route.extend(namespaces)
        return node

    @classmethod
    def get_entities(
        cls,
        search_dir: PathLike,
        recursive: bool = True
    ) -> list[Entity]:
        """Add subspace node to the tree based on entity.

        Args:
            entity (Entity): [description]
        """
        search_path: Path = Path(search_dir).resolve()
        entities: list[Entity] = []
        if recursive:
            # FIXME: replace this with Path.walk (python 3.12)
            # https://docs.python.org/3/library/pathlib.html
            for root, dirs, files in os.walk(search_path):
                for path in [*dirs, *files]:
                    child = Path(root, path).resolve()
                    if cls.is_entity(child):
                        entities.append(child)
        else:
            for child in search_path.iterdir():
                if cls.is_entity(child):
                    entities.append(child)
        return entities

    @staticmethod
    def is_entity(path: Path) -> bool:
        """Return True if path is a valid entity.

        Args:
            path (Path): [description]

        Returns:
            bool: [description]
        """
        path = path.resolve()
        if path.is_dir():
            is_subspace = Path(path, 'Subspace.yml').is_file()
            is_void = 'Void' in path.name.split("_")
            return is_subspace or is_void
        else:
            not_marker = path.name != 'Subspace.yml'
            return not_marker

    def render_tree(self):
        """Render the tree as rich str.

        Returns:
            [type]: [description]
        """

        label_dict = {
            "📁": "direcotry subspace\nwhich contains Subspace.yml.",
            "📄": "file subspace\nwhich refers to a leaf file.",
            "💿": "phantom | unknown subspace\nwhich has no entity.",
            "⛔": "virtual subspace\nwhich have not created yet."
        }

        return Grid(
            self.format(repr="{node.data.ui_name}", title="(Root)"),
            Instruction(label_dict)
        )

    def render_table(self):
        """List subspaces in table as rich str.

        Returns:
            [type]: [description]
        """

        table = Table(
            "Subspace",
            "Comments",
            "Entities"
        )

        for node in self:
            subspace: Subspace = node.data
            table.add_row(
                subspace.ui_name,
                "\n".join(subspace.ui_comments),
                "\n".join(subspace.ui_entities)
            )
        return table

    def draw_graph(
        self,
        output_dir: PathLike,
        reveal_when_success: bool = True
    ):
        """Draws the graph and export html file.

        Args:
            output_dir (PathLike): [description]
            reveal_when_success (bool, optional): [description]. Defaults to True.
        """
        node_dict = {
            ".": {
                "name": "(Root)",
                "content": "&nbsp(Root)&nbsp",
                "color": "#BEBEBE",
                "size": 20,
                "level": 0,
                "border_width": 0,
            }
        }
        for node in self:
            subspace: Subspace = node.data
            parent: Subspace = node.parent.data if node.parent else None

            if (subspace.type == SubspaceType.PHANTOM):
                color = "#939393"
            elif (subspace.type == SubspaceType.DIRECTORY):
                color = "#F6B330"
            else:
                color = "#585858"

            node_dict[node.path] = {
                "name": subspace.html_name,
                "parent": node.parent.path if parent else ".",
                "content": "<br>".join(subspace.html_entities),
                "color": color,
                "size": 5+subspace.flow,
                # "level": node.depth,
                "border_width": 0,
                "edge_width": 5+subspace.flow,
                "edge_color": '#3F3F3F',
            }

        draw_graph(
            node_dict,
            bgcolor="#111112",
            font_color="#F4F4F4",
            output_dir=output_dir,
            reveal_when_success=reveal_when_success
        )




from pathlib import Path
import networkx as nx
from pyvis.network import Network

from omoospace.utils import PathLike, reveal_in_explorer


def draw_graph(
    node_dict: dict,
    output_dir: PathLike = ".",
    bgcolor: str = "#ffffff",
    font_color: bool = False,
    reveal_when_success: bool = True
):
    G = nx.Graph()
    for key, node in node_dict.items():
        G.add_node(
            key,
            label=node.get('name'),
            title=node.get('content'),
            level=node.get('level'),
            size=node.get('size'),
            color=node.get('color'),
            borderWidth=node.get('border_width')
        )
        if (node.get('parent')):
            G.add_edge(
                node.get('parent'),
                key,
                weight=node.get('edge_width'),
                color=node.get('edge_color')
            )
    net = Network(
        "100vh",
        notebook=True,
        bgcolor=bgcolor,
        font_color=font_color,
        cdn_resources='remote'
    )
    net.from_nx(G)
    output_path = Path(output_dir)
    output_path.mkdir(parents=True, exist_ok=True)
    structure_filepath = Path(output_path, "Structure.html").resolve()
    net.show(str(structure_filepath))

    if reveal_when_success:
        reveal_in_explorer(structure_filepath)




from pathlib import Path
import os
from enum import Enum
from typing import TypedDict, Union

Structure = dict[str, "Structure"]

Entity = Path

Item = Path

Route = list[str]

PathLike = Union[str, os.PathLike]


class Work(TypedDict):
    name: str
    paths: list[str]


class Creator(TypedDict):
    name: str
    email: str
    role: str
    website: str


class Plugin(TypedDict):
    name: str
    version: str


class Software(TypedDict):
    name: str
    version: str
    plugins: list[Plugin]


class OmoospaceInfo(TypedDict):
    name: str
    description: str
    creators: list[Creator]
    softwares: list[Software]
    works: list[Work]


class SubspaceInfo(TypedDict):
    name: str
    description: str
    creators: list[Creator]


class PackageInfo(TypedDict):
    name: str
    description: str
    version: str
    creators: list[Creator]


class OmoospaceStructure(TypedDict):
    Contents: dict
    SourceFiles: dict


class SubspaceType(Enum):
    DIRECTORY = "directory"
    FILE = "file"
    PHANTOM = "phantom"



import os
from pathlib import Path
from zipfile import ZipFile


from omoospace.exceptions import NotFoundError
from omoospace.types import Item, PathLike
from omoospace.common import console, yaml


class Package:
    """The class of omoospace package.

    A package class instance is always refer to a existed package directory, not in ideal. 

    Attributes:
        name (str): Package's name.
        description (str): Package's description.
        version (str): Package's version.
        creators (list[Creator]): Creator list.
        root_path (Path): Root path.
    """

    def __init__(self, detect_dir: PathLike):
        """Initialize package .

        Args:
            detect_dir (PathLike): [description]

        Raises:
            NotFoundError: [description]
            NotFoundError: [description]
        """
        package_path = Path(detect_dir).resolve()
        if (package_path.suffix == ".zip"):
            with ZipFile(package_path, 'r') as zip:
                try:
                    with zip.open('Package.yml') as file:
                        package_info = yaml.load(file)
                except:
                    raise NotFoundError("package", detect_dir)
        else:
            package_info_path = Path(package_path, 'Package.yml')
            if package_info_path.exists():
                with package_info_path.open('r', encoding='utf-8') as file:
                    package_info = yaml.load(file)
            else:
                raise NotFoundError("package", detect_dir)

        self.root_path = package_path
        self.name = package_info.get('name')
        self.description = package_info.get('description')
        self.version = package_info.get('version')
        self.creators = package_info.get('creators')

    # TODO: omoospace also has is_package_item. Keep one only.
    @staticmethod
    def is_package_item(path: Path) -> bool:
        not_marker = path.name != '.subspace'
        not_package_info = path.name != 'Package.yml'
        return not_marker and not_package_info

    @property
    def items(self) -> list[Item]:
        """list[Item]: The list of Item objects."""
        items: list[Item] = []
        for root, dirs, files in os.walk(self.root_path):
            for path in files:
                child = Path(root, path).resolve()
                if self.is_package_item(child):
                    items.append(child)
        return items



import click
from pathlib import Path
from InquirerPy import inquirer
from InquirerPy.base import Choice
from prompt_toolkit.validation import ValidationError, Validator
from omoospace.ui import Board, Info

from omoospace.utils import format_name
from omoospace.omoospace import Omoospace
from omoospace.commands.common import inquirer_style, to_omoospace
from omoospace.setting import OmoospaceTemplate, Setting
from omoospace.common import console


@click.command('create', help="Create new omoospace.")
@click.option("-d", "--dst", default='.',
              type=click.Path(exists=True, file_okay=False))
def cli_create(dst: str):
    class OmoospaceNameValidator(Validator):
        def validate(self, document):
            # Check if is empty
            if not len(document.text) > 0:
                raise ValidationError(
                    message="Name cannot be empty.",
                    cursor_position=document.cursor_position,
                )

            omoospace_name = format_name(document.text)
            omoospace_path = Path(dst, omoospace_name).resolve()

            # Check if omoospace exists
            if (omoospace_path.exists()):
                raise ValidationError(
                    message="'%s' already exists in dstination directory, "
                    "change the name or the dstination." % omoospace_name,
                    cursor_position=document.cursor_position,
                )

    templates = Setting().omoospace_templates
    template_choices = [
        Choice(name="*Empty", value=-1),
        *[Choice(
            name="%s (%s)" % (template.name, template.description),
            value=id
        ) for id, template in enumerate(templates)]
    ]

    template_id: int = inquirer.select(
        message="Choose a template",
        choices=template_choices,
        style=inquirer_style,
    ).execute()

    while True:
        if template_id >= 0:
            template = templates[template_id]
        else:
            template = OmoospaceTemplate()
        console.print(template.tree.render_tree())
        confirm_id: int = inquirer.select(
            message="Choose the same again to confirm",
            choices=template_choices,
            default=template_id,
            long_instruction="Press Enter key to confirm",
            style=inquirer_style,
        ).execute()
        if confirm_id == template_id:
            break
        else:
            template_id = confirm_id

    omoospace_name: str = inquirer.text(
        message="Enter Omoospace name",
        validate=OmoospaceNameValidator(),
        style=inquirer_style
    ).execute()
    omoospace_path = Path(dst, format_name(omoospace_name)).resolve()

    description = inquirer.text(
        message="A brief of the Omoospace",
        default="An omoospace for creation works",
        style=inquirer_style
    ).execute()

    omoospace_info = {
        "description": description
    }

    # Confirm information

    console.print(Board(
        Info("Name", "%s [dim](%s)[/dim]" %
             (omoospace_name, omoospace_path)),
        Info("Description", description),
        Info("Directory Tree", template.tree.render_tree()),
        title="Create Omoospace Form"
    ))
    is_confirm = inquirer.confirm(
        message="Confirm",
        default=True,
        style=inquirer_style
    ).execute()
    if is_confirm:
        omoospace = Omoospace.create(
            create_dir=dst,
            name=omoospace_name,
            info=omoospace_info,
            structure=template.structure
        )
        to_omoospace(omoospace.root_path)
        console.print("Start create something! 🚀", style="cheer")



import click
from pathlib import Path
from InquirerPy import inquirer
from InquirerPy.validator import EmptyInputValidator
from prompt_toolkit.validation import ValidationError, Validator
from omoospace.ui import Board, Info, Table, Transfer

from omoospace.utils import format_name
from omoospace.commands.common import collect_item_prompt, option_omoospace, get_omoospace, inquirer_style
from omoospace.common import console


@click.command('export', help="Export new package.")
@option_omoospace()
@click.option(
    "-d", "--dst", type=click.Path(),
    help="The output path of Package"
)
def cli_export(omoospace, dst):
    omoospace = get_omoospace(omoospace)

    package_items: list[Path] = collect_item_prompt(
        omoospace.root_path, omoospace.is_omoospace_item)

    package_name = inquirer.text(
        message="Name this package",
        validate=EmptyInputValidator("Name cannot be empty"),
        default=omoospace.name,
        style=inquirer_style
    ).execute()

    package_description = inquirer.text(
        message="Enter a brief description of this package",
        default="An omoospace package for sharing",
        style=inquirer_style
    ).execute()
    package_version = inquirer.text(
        message="Enter the verison of this package",
        default="0.1.0",
        style=inquirer_style
    ).execute()

    package_info = {
        "description": package_description,
        "version": package_version
    }

    ui_items = Table("Name", "Path")
    for item in package_items:
        _abspath = str(Path(omoospace.root_path, item))
        _name = Path(_abspath).parts[-1]
        ui_items.add_row(_name, _abspath)

    export_path = Path(dst).resolve() if dst \
        else Path(omoospace.root_path, 'StagedData', 'Packages').resolve()
    package_path = Path(export_path, format_name(package_name)).resolve()
    console.print(Board(
        Info("Name", "%s [dim](%s)[/dim]" %
             (package_name, package_path)),
        Info("Description", package_description),
        Info("Version", package_version),
        Info("Items", ui_items),
        title="Export Package Form"
    ))

    is_confirm = inquirer.confirm(
        message="Sure",
        default=True,
        style=inquirer_style
    ).execute()
    if is_confirm:
        omoospace.export_package(
            items=package_items,
            export_dir=export_path,
            name=package_name,
            info=package_info
        )
        console.print("Successfully export! 📦", style="cheer")



from rich.theme import Theme
from rich.console import Console
from ruamel.yaml import YAML

yaml = YAML()
yaml.indent(sequence=4, offset=2)

custom_theme = Theme({
    "bright": "#ffd700",
    "dim_bright": "#808080",
    "dim": "#808080",
    "warning": "bold #ffffff on #ffd700",
    "danger": "bold #ffffff on red",
    "cheer": "bold #ffffff on blue"
})
console = Console(theme=custom_theme)




from typing import Iterable, Optional, Union
from rich.align import AlignMethod
from rich.padding import PaddingDimensions
from rich.table import Table as R_Table, Column
from rich.panel import Panel
from rich.console import Group, RenderableType
from rich import box
from rich.columns import Columns
from rich.text import TextType
from omoospace.types import Creator


class Card(Panel):
    def __init__(self, renderable, title=None, width=None):
        super().__init__(
            renderable,
            title=title,
            width=width
        )


class Board(Panel):
    def __init__(self, *renderables, title=None) -> None:
        super().__init__(
            Group(*renderables),
            title=title,
            expand=True,
            padding=(1, 3),
            box=box.ROUNDED
        )


class Info(R_Table):
    def __init__(self, key, value) -> None:
        super().__init__(
            box=box.ROUNDED,
            show_edge=False,
            show_header=False,
            title=key,
            title_style="b i",
            title_justify="left",
            padding=(0, 0, 1, 0),
            expand=True
        )
        self.add_column("")
        self.add_row(value)


class Instruction(Card):
    def __init__(self, label_dict: dict[str, str]) -> None:
        labels = ["[dim]%s --- %s" % (key, value)
                  for key, value in label_dict.items()]
        super().__init__(
            "\n\n".join(labels), title='instruction'
        )


class Table(R_Table):
    def __init__(self, *headers: list[Union[Column,dict,str]], title=None, rows: list[list] = []):
        headers = [Column(**header)if isinstance(header, dict)
                   else Column(header) for header in headers]
        super().__init__(
            *headers,
            title=title,
            box=box.ROUNDED,
            header_style="i",
            expand=True
        )
        for row in rows:
            self.add_row(*row)


class Grid(R_Table):
    def __init__(self, *columns, vertical="top", expand=True):
        # grid = Columns(renderables)
        super().__init__(
            box=box.SIMPLE,
            show_header=False,
            show_edge=False,
            padding=(0, 0),
            expand=expand
        )
        renderables = []
        for column in columns:
            if isinstance(column, dict):
                ui = column.get('ui')
                kwargs = column.get('kwargs')
            else:
                ui = column
                kwargs = {
                    "header": "",
                    "vertical": vertical,
                }
            self.add_column(**kwargs)
            renderables.append(ui)
        self.add_row(*renderables)


class Transfer(Grid):
    def __init__(self, left_list, right_list, left_header="Left", right_header="Right"):
        super().__init__(
            Panel(
                "\n".join(["? %s" % item for item in left_list]),
                title=left_header,
                style="bright",
                border_style="white",
                padding=(1, 4)
            ),
            {"ui": "=>", "kwargs": {"justify": "center", "vertical": "middle"}},
            Panel(
                "\n".join(["√ %s" % item for item in right_list]),
                title=right_header,
                style="dim",
                border_style="white",
                padding=(1, 4)
            ),
            vertical="middle",
            expand=False
        )


class CreatorProfile(Panel):
    def __init__(self, creator: Creator):
        super().__init__(
            Group(
                Info("[bright]Name[/bright]", creator.get("name")),
                Info("[bright]Email[/bright]", creator.get("email")),
            ),
            title="CREATOR",
            width=50,
            expand=False,
            padding=(1, 3)
        )


class ItemList(Columns):
    def __init__(self, items: list[str]) -> None:
        super().__init__([Card(item) for item in items], expand=False)



import click
import importlib

from omoospace.commands.create import cli_create
from omoospace.commands.set import cli_set
from omoospace.commands.to import cli_to
from omoospace.commands.export import cli_export
from omoospace.commands.show import cli_show
module_import = importlib.import_module("omoospace.commands.import")
cli_import = module_import.cli_import

# TODO: remove null in yaml
# TODO: more manage of templates
# TODO: docs


@click.group()
def cli():
    pass


cli.add_command(cli_set)
cli.add_command(cli_create)
cli.add_command(cli_show)
cli.add_command(cli_to)
cli.add_command(cli_export)
cli.add_command(cli_import)

if __name__ == "__main__":
    cli()
